!
!  Module      : ogprep_mod
!
! Authors     : Vladimir Yershov (MSSL)
!          
!
! Description : Alignes OM detector (grism) spectra along y-coordinate
! 
! Comments    : 
!
! $Id: module.f90,v 1.4 1997/12/18 16:01:10 gvacanti Exp $
! 
! History : Created  : V0.1 October 2002 
! VNY 06.11.2003 Parameter for the output undistorted image made non-mandatory
! VNY 07.11.2003 The background around the rotated image is set to -1
! VNY 11.11.2003 the code is tiden up
! 25.11.2003 (VNY) distortion correction improvement
!
!*************************************************************************** 
module omgprep_mod

  use param
  use cal
  use dal
  use sasdal
  use caloaldefs

  use errorHandling
  use sla
  use readModes_mod
  use allocation_mod 
  use utility_mod
!  use sourcematch_mod
!  use astromcorrect_mod    ! For doing the astrometry correction
  use omconstants_mod

implicit none
 
  private

   public omgprep    

contains

!---------------------------------------------------------------------
subroutine omgprep
!---------------------------------------------------------------------

    type(DataSetT) :: inDataSetH                                       ! Input dataset handle.
!    type(DataSetT) :: sourceListDataSetH                              ! Source List dataset handle.
    type(DataSetT) :: PPSOSWDataSetH                                  ! PPS OSW dataset handle.
!    type(DataSetT) :: smoothDataSetH                                  ! hadle for the smoothed output dataset
    type(DataSetT) :: undistDataSetH                                  ! hadle for the undistorted output dataset
    type(ArrayT) :: inArrayH                                          ! Input array handle.
    type(ArrayT) :: qualArrayH                                        ! Quality Array handle.
    type(ArrayT) :: PPSOSWArrayH                                      ! PPS OSW array handle.
    type(ArrayT) :: undistArrayH                                      ! smooth array handle
    type(AttributableT) :: sourceH                                    ! Source attributable handle.
    type(AttributableT) :: objectH                                    ! Object attributable handle.
    type(BlockT) :: blockH                                            ! Block handle.

    type (SpacecraftAttitudeType) attitude

    character(len=255) :: set                                         ! OM image intermediate filename
    character(len=255) :: PPSOSWSet                                   ! PPS OM OSW FITS sky image filename (rotated).
    character(len=255) :: undistSet                                   ! filename for the undistorted image    
    character(len=20):: rastring, decstring   ! strings for RA and DEC
    character(len=1) :: cs                    ! sign of dec +/-
    character(len=10)           :: datamode   ! datamode attribute
    character(len=6)            :: omFilter   ! the OM filter

    integer                    :: dim(2), nx, ny     ! size of the image
    integer                    :: xBinning
    integer                    :: yBinning
    integer             :: nxout, nyout        ! size of the output image
    integer             :: nxdist, nydist        ! size of the output udistorted image
    integer :: i, j, k                           ! counters
    integer             :: xbin, ybin         ! binning factors
    integer             :: xsize, ysize       ! size of the image
    integer             :: x0, y0             ! offset of the image
    integer             :: idmsf(4), ihmsf(4)  
    integer                     :: OM_MODE    ! the om mode

    real(kind=single), pointer :: oswImage(:, :)
    real                       :: refDX
    real                       :: refDY
    real                       :: rotatedRefDX
    real                       :: rotatedRefDY
    real(kind=double)  :: itrans(6)     ! its inverse

    ! sky coordinate image variables
    real(kind=single), pointer  :: outImage(:,:)       ! The output image
    real(kind=single), pointer :: undistImageP(:,:)     ! The undistorted output image
    real(kind=single), dimension(:, :), allocatable :: undistImage 
    real(kind=single)          :: exposureTime         ! The exposure time for an input image
    real(kind=double)  :: trans(6) ! rotation transformation matrix
                         
    real(kind=double)  :: rotationCorrection = 90.0_double! Correction for coordinate error
    real(kind=double)  :: dXbinning, dYbinning
    real(kind=double)  :: sinRollAngle, cosRollAngle
   
    real(kind=single)  :: refx, refy         ! SRR added 4/10/00
    real(kind=double)  :: xmax,xmin,ymax,ymin            
    real(kind=double) :: ra_scx, dec_scx, pa_scx  ! The ra, dec and pa of the spacecraft     

    real(kind=double) :: ra_point, dec_point        ! position of the OM tangent point
    real(kind=double) :: rollangle                 ! the roll angle
    real(kind=double) :: time                        
    real(kind=double) :: attmatrix(1 : 3, 1 : 3)    ! rotation matrix
    real(kind=double) :: attmatrix1(0 : 2, 0 : 2)    ! rotation matrix

    real(kind=double) :: att_invbor(3, 3)           !matrix for product of attitude*boresight
    real(kind=double) :: ompoint(3)=(/1.0_double, 0.0_double, 0.0_double/)         !om boresight vector in OM frame
    real(kind=double) :: pointvec(3)        ! s/c pointing
    real(kind=double) :: pointvec2(3)       
    real(kind=double) :: euler_axis(3)              ! euler axis for rotation
    real(kind=double), pointer :: boremat(:,:)       ! boresight offset matrix
    real(kind=double) :: inv_boremat(3, 3)          ! matrix for inverse of boremat
    real(kind=double) :: omvec(3)                   ! om pointing vector
    real(kind=double) :: rotmat(3, 3)               ! rotation matrix
    real, pointer     :: deltax(:), deltay(:)       ! distortion offsets
    real(kind=double) :: r1                         ! modulus
    real(kind=double) :: currentEpoch
    real               :: rawx(0 : 0), rawy(0 : 0)  ! CAL coords
    real(kind=double)           :: lambda           ! wavelength for CAL
    real(kind=double)           :: plateScale       ! the platescale
    real :: white, black!, grey, grey2
    real(kind=single) :: nullValue
   integer(kind=int32) :: windowX0, windowX1, windowY0, windowY1 !+   ! Coordinates of the OSW origin
   integer(kind=int32) :: windowDX, windowDY !+
   integer(kind=int32) :: winX0, winY0, winX1, winY1, winDX, winDY
   character(len=1) :: ch1, ch2, ch3, ch4, ch5
   character(len=255) :: window0 

    logical(kind=Bool) :: binBPE                    ! BPE binning on?
    logical(kind=bool) :: produceUndistortedImage ! condition for producin an auxiliar output image
!    integer                       :: istat              ! status for memory allocation
    integer(kind=int32)           :: filterIdInt
    real(kind=single) :: averImage
    real(kind=single) :: sum !, cosPA, sinPA
    character(len=255) :: detsec, detsize
    integer(kind=int8), dimension(:, :), pointer :: qualArrayP        ! Quality array pointer.
    integer(kind=int8), dimension(:, :), pointer :: outQualityArrayP  ! Quality-2 array pointer.
    logical(kind=bool), dimension(:, :), allocatable :: qualArray     ! Quality array.
    integer(kind=int8), dimension(:, :), allocatable :: outQualArray 
    logical(kind=bool) :: hasQUALITY
    type(ArrayT) :: outQualityArrayH
    integer(kind=int32) :: allocErr                                   ! Allocate error status.
    character(len=255) :: attrValue
    integer(kind=int32) :: int32Value
    real(kind=single)  :: real32Value
    nullValue = real(-99999)
    produceUndistortedImage = .false.
    !*************************************
    ! Get the command-line parameters
    !*************************************
    set = stringParameter("set")
    PPSOSWSet = stringParameter("outset")
    !smoothSet = stringParameter("smoothset")
    undistSet = stringParameter("undistset")
    if (undistSet .ne. "") then
       produceUndistortedImage = .true.
    endif

!!******************************************** 

    ! the image contains the information on the science window
    inDataSetH = DataSet(set, modify) 
    inArrayH = array(inDataSetH, "PRIMARY", modify)
    exposureTime = real32Attribute(inDataSetH, "EXPOSURE") ! Image exposure time
!=================================

    hasQUALITY= .false.
    if (hasBlock(inDataSetH, "QUALITY")) then
       hasQUALITY= .true.
    endif

    xbin = int32Attribute(inDataSetH, "BINAX1")      ! binning factors
    ybin = int32Attribute(inDataSetH, "BINAX2")
 
    if (xbin .ne. ybin) then
        call fatal("diffBin", "This version cannot deal with different binning factors in x and y")
    end if

    binBPE = booleanAttribute(inDataSetH, "BINBPE")

    ! Calculate image binning.
    if (xbin .eq. 0) xbin=1
    if (ybin .eq. 0) ybin=1
    xBinning = xbin
    yBinning = ybin

    if (binBPE) then
        xBinning = 2.0
        yBinning= 2.0
    end if

    dXbinning = dble(xBinning)  ! Need the double values of xBinning and 
    dYbinning = dble(yBinning)  ! yBinning fairly frequently
 
    !-(origin of the OSW)

       if (hasAttribute(inArrayH,"WINDOWX0")) then
          windowX0= int32Attribute(inArrayH, "WINDOWX0")
          windowY0= int32Attribute(inArrayH, "WINDOWY0")
          windowDX= int32Attribute(inArrayH, "WINDOWDX")
          windowDY= int32Attribute(inArrayH, "WINDOWDY")
          winX0=windowX0/xBinning
          winY0=windowY0/yBinning
          winDX=windowDX/xBinning
          winDY=windowDY/yBinning
          winX1=winX0+winDX-1.0
          winY1=winY0+winDY-1.0
       else
          if (hasAttribute(inArrayH,"DETSEC")) then
             window0 = stringAttribute(inArrayH, "DETSEC")
             read (window0, '(A1, I4, A1, I4, A1, I4, A1, I4, A1)')  &
                  ch1,windowX0, ch2, windowX1, ch3,  windowY0, ch4, windowY1, ch5
             winX0=windowX0
             winY0=windowY0
             winX1=windowX1
             winY1=windowY1
             winDX=windowX1-windowX0+1
             winDY=windowY1-windowY0+1
          else
             dim = dimensions(inArrayH)   ! Get the size of the image 
             winX0=0
             winY0=0
             winDX = dim(1)
             winDY = dim(2)
             winX1= winDX-1.0
             winY1= winDY-1.0
          endif
       endif

    omFilter = stringAttribute(inDataSetH, "FILTER")
    if (trim(omFilter) .eq. "Grism1") omFilter="GRISM1"
    if (trim(omFilter) .eq. "Grism2") omFilter="GRISM2"

    if (trim(dataMode) .eq. "IMAGING") then 
       OM_MODE = IMAGE
    else
       OM_MODE = FAST
    end if

    !***********************************************************
    ! Read in WINDOWXO (x0) & WINDOWY) (y0) and the image size
    !***********************************************************
    call readModes(inDataSetH, x0, y0, xsize, ysize)  

    oswImage => real32array2data(inArrayH)
    dim = shape(oswImage)
    nx = dim(1)
    ny = dim(2)   
    allocate (qualArray(0 : nx-1, 0 : ny-1), stat = allocErr)
    call CheckAllocation(allocErr, "qualArray")
    ! Get a handle and pointer on the quality array and extract the quality data.
    qualArray=.false.
    if (hasQUALITY) then
       qualArrayH = array(inDataSetH, "QUALITY", read)
       qualArrayP => int8Array2Data(qualArrayH)
       !outQualArray=qualArrayP
       where (qualArrayP .ne. 0) 
          qualArray = .true.
       elsewhere
          qualArray = .false.
       end where
    endif

    !***********************************************************
    ! Check the source shapes 
    !***********************************************************
    black = minval(oswImage)
    white = maxval(oswImage)
    averImage=0.0
    sum=0.0

    do i=0,nx-1
       do j=0,ny-1
          averImage=averImage+oswImage(i,j)
          sum=sum+1.0
       enddo
    enddo

    if (sum .gt. 0.0) then
       averImage=averImage/sum
    endif

    !*******************************************************
    ! get the astrometry info keywords 
    !*******************************************************
    ra_scx = real64Attribute(inDataSetH, "RA_SCX") * degreesToRadians   ! Spacecraft RA in radians
    dec_scx = real64Attribute(inDataSetH, "DEC_SCX") * degreesToRadians ! Spacecraft dec in radians
    pa_scx = real64Attribute(inDataSetH, "PA")        ! Spacecraft PA in degrees

    !********************************************
    ! Transform the source position angles
    ! to the equatorial coordinate system
    !*******************************************
    pa_scx = pa_scx * degreesToRadians
    rollAngle = pa_scx

    !************************************
    ! Get the filter Id integer value
    !************************************
    filterIdInt = filterIdToInt(omFilter)
    !*********************
    ! Set the CAL state
    !*********************
    call CAL_setState(OM, ccdModeId=OM_MODE, filterId = filterIdInt)
    call CAL_omGetPlateScale(lambda, plateScale)  ! Get the plate-scale (arcsecs per pixel) 
    !plateScale=0.476513
    !***********************************************************
    ! Get the number of seconds since the mission reference time
    !***********************************************************
    time =  MissionTime(inDataSetH)

!    currentEpoch = real32Attribute(sourceListDataSetH, "EQUINOX")
!***
    currentEpoch=2000.0
!***
    !*********************************************************
    ! Need the current epoch for the astrometry correction
    !*********************************************************
    currentEpoch = 1998.0_double + time / (365.25_double * 86400.0_double)
    call CAL_GetBoresightMatrix(time, boremat) ! boresight offset matrix
    !*******************************************************************
    ! need inverted boresight matrix which is obtained via the transpose. 
    !********************************************************************
    inv_boremat = TRANSPOSE(boremat)
    attitude = SpaceCraftAttitudeType(ra_scx, dec_scx, rollangle)
    rollangle=0.0

    ! Visual grism rollangle (Grism2)
    if (omFilter .eq. "Grism2  " .or. omFilter .eq. "GRISM2  ") rollangle=-5.6
    if (omFilter .eq. "Grism1 " .or. omFilter .eq. "GRISM1  ") rollangle=6.3

    !*******************************************************
    ! This routine requires a matrix of size 0:2, 0:2, but
    ! after that we want it to be of size 1:3, 1:3
    !******************************************************
    call OAL_toAttitudeMatrix(attitude, attmatrix1)
    
    forall( i = 1 : 3) 
        forall(j = 1 : 3) attmatrix(i, j) = attmatrix1(i - 1, j - 1)
    end forall

    !**********************************************
    ! rotationCorrection corrects for OM axes definition.
    !**********************************************
    rotationCorrection =0.0
    rotationCorrection = rotationCorrection * degreesToRadians 

    !******************************************************************************
    ! converts RA and DEC from spherical cordinates to cartesian coordinates vector
    !*****************************************************************************
    call sla_dcs2c(ra_scx, dec_scx, pointvec)

    !********************************************************
    ! normalize the vector pointvec and return in pointvec2 -
    ! r1 holds the mdulus of the vector. pointvec2 will store
    ! the unit-vector in the direction of pointvec
    !********************************************************
    call sla_dvn(pointvec, pointvec2, r1)

    !--------------------------------------------------------------------------------------------------------
    pointvec(1) = -pointvec2(1)
    pointvec(2) = -pointvec2(2)
    pointvec(3) =  pointvec2(3)
    !--------------------------------------------------------------------------------------------------------
    euler_axis = -rollangle * pointvec        !leave in (SRR)

    ! convert euler axis vector to a rotation matrix
    call sla_dav2m(euler_axis, rotmat) 

    ! initialise output vector
    omvec = 0.0_double

!**********************************************************************************************
!  To get boresight correction, apply attitudematrix*(boresightmatrix)^(-1)*ompoint(=1,0,0)
!  Boresight * vector in SC frame gives vector in OM frame so boresight^(-1)*vector in OM frame gives 
!  vector in SC frame.  Then attitude * vector in S/C frame gives vector in Equatorial Reference Frame.
!
!  This creates omvec, the om boresight vector measured in the ERF
!  Note this process is applied here by first doing attmatrix*inv_boremat, then multiplying the 
!  ompoint vector by the product and decomposing to RA and DEC. The ompoint vector 
!  is the OM boresight as measured in the OM reference frame
!*********************************************************************************************
    att_invbor = matmul(attmatrix, inv_boremat)
    omvec = matmul(att_invbor, ompoint)
    !************************************************************************
    ! converts the omvec cartesian vector to ra and dec spherical coordinates
    ! and check RA lies within sensible range
    ! ra_point and ra_dec refer to the pointing direction of the instrument
    !************************************************************************
    ra_point=0.0
    dec_point=0.0
    call sla_dcc2s(omvec, ra_point, dec_point)

    ! Protection against the NaN values of ra ad dec
    if (ra_point .eq. nullValue .and. dec_point .eq. nullValue) then
       ra_point=0.0
       dec_point=0.0
    endif
    if (ra_point .lt. 0.0_double) ra_point = ra_point + 2.0_double * pi

    write(errstr, '(a, f8.4)') "RA of OM boresight: ", ra_point * radiansToDegrees
    call message(AppMsg, Verbose)
    write(errstr, '(a, f8.4)') "DEC of OM boresight: ", dec_point * radiansToDegrees
    call message(AppMsg, Verbose)

    deltax => AllocateSingle(0, 0, "deltax")
    deltay => AllocateSingle(0, 0, "deltay")
 
! Test VNY 17.07.2003
!    do i=1000,1010!
!
!       rawx(0) = real(i)
!       rawy(0) = real(i)!
!
!       call CAL_omDistortion(rawx, rawy, deltax, deltay)
!       write(*,*) "DIST L800 i=", i, " deltax=", deltax, " deltay=", deltay
!    enddo
!---

    !****************************************************************
    !  NOW CREATE THE OUTPUT PPS PRODUCT OM OSW SKY FITS IMAGE. 
    !  This is done by resampling the detector image on a sky grid. 
    ! the transformation is currently a simple rotation
    ! Define the transformation terms (note the offsets will change)
    !****************************************************************
    rollangle=rollangle * degreesToRadians 
    trans(1) =  0.0_double
    trans(2) =  dcos(-rollangle)
    trans(3) =  dsin(-rollangle)
    trans(4) =  0.0_double
    trans(5) = -dsin(-rollangle)
    trans(6) =  dcos(-rollangle)
    sinRollangle = dsin(rollangle)
    cosRollangle = dcos(rollangle)
    !*******************************************
    ! create the output source file
    !*******************************************
    PPSOSWDataSetH = DataSet(trim(PPSOSWSet), create)
    !smoothDataSetH = DataSet(trim(smoothSet), create)
    if (produceUndistortedImage) undistDataSetH = DataSet(trim(undistSet), create)

!  determine the image boundaries for the undistorted image
    itrans(1)=0.0_double
    itrans(2)=1.0_double
    itrans(3)=0.0_double    
    itrans(4)=0.0_double
    itrans(5)=0.0_double
    itrans(6)=1.0_double
! (the rotation matrix itrans is set to the zero rotation angle)

    call  GetImageBoundaries(nx, ny, real(x0), real(y0), xBinning, yBinning, &
                             itrans, xMin, xMax, yMin, yMax)    
    nxdist=xmax-xmin
    nydist=ymax-ymin

!--- VNY the undistorted image probably is better to have the same dimensions as the original one
!    albeit it can lead to some information loss at the edges
    nxdist=nx
    nydist=ny
!---
!**************************************************************************
!  Test creating the undistorted image (21.07.03)
!**************************************************************************
    dim(1) = nxdist     ! The dimensions of the output image
    dim(2) = nydist
    ! copy the header across
    ! VNY 07.11.2003
    if (produceUndistortedImage) then 
       call copyAllAttributes(attributable(inDataSetH), attributable(undistDataSetH))
       undistArrayH    = addArray(undistDataSetH, "PRIMARY",  Real32, dim, "Counts", "IMAGE")  ! For the image array

       sourceH = attributable(inDataSetH)
       objectH=attributable(undistArrayH)
       call addAttributes(objectH, sourceH)

       call setAttribute(undistDataSetH, "FRAMTIME", 21)
       ! add content keyword
       call setAttribute(undistDataSetH, "CONTENT", "OM OSW FITS UNDISTORTED IMAGE", &
            "Content of FITS file")
       call setAttribute(undistDataSetH, "HDUCLASS", "OGIP    ", "Format conforms to OGIP/GSFC conventions")

       blockH=block(inDataSetH, "PRIMARY", read)
       sourceH=attributable(blockH)

       call addAttributes(objectH, sourceH)

       call setAttribute(objectH, "HDUCLASS", "OGIP    ", "Format conforms to OGIP/GSFC conventions")
    
       blockH=block(inDataSetH, "QUALITY", read)
       call copyBlock(undistDataSetH, blockH)
       blockH=block(inDataSetH, "MODES", read)
       call copyBlock(undistDataSetH, blockH)
    endif ! produceUndistortedImage

    !**********************************************************************
    ! Determine the bounding x and y (xmin, xmax, ymin, ymax) coordinates 
    ! of the image
    !*********************************************************************
    call  GetImageBoundaries(nx, ny, real(x0), real(y0), xBinning, yBinning, &
                             trans, xMin, xMax, yMin, yMax)

    ! invert transformation matrix for later use
    call invert_trans(trans, itrans)    ! need inverted transformation matrix later to rotate image

    nxout = xmax - xmin                  ! determine image dimensions
    nyout = ymax - ymin

    !***********************************************************************************
    ! set up the output image with the appropriate size (xmax,ymax already referenced to 
    ! origin of containing (output) box so these equal the output box dimensions).
    !**********************************************************************************
    dim(1) = nxout     ! The dimensions of the output image
    dim(2) = nyout

    ! copy the header across
    PPSOSWArrayH    = addArray(PPSOSWDataSetH, "PRIMARY",  Real32, dim, "Counts", "IMAGE")  ! For the image array

    sourceH = attributable(inDataSetH)
    objectH=attributable(PPSOSWArrayH)
   
    !call addAttributes(objectH, sourceH)
    ! add content keyword
    call setAttribute(PPSOSWDataSetH, "CONTENT", "OM OSW FITS ROTATED IMAGE", &
         "Content of FITS file")
    ! Add attribute HDUCLASS
    call setAttribute(PPSOSWDataSetH, "HDUCLASS", "OGIP    ", "Format conforms to OGIP/GSFC conventions")

    blockH=block(inDataSetH, "PRIMARY", read)
    sourceH=attributable(blockH)
    !call addAttributes(objectH, sourceH)
    call setAttribute(objectH, "HDUCLASS", "OGIP    ", "Format conforms to OGIP/GSFC conventions")

    write (detsec, '(a1,i4,a1,i4,a1,i4,a1,i4,a1)') "[",0, ":", nxout-1, &
         & ",",0, ":", nyout-1, "]" 
    if (xbin .ne. 0. .and. ybin .ne. 0.) then
       write (detsize,'(a6,i4,a6,i4,a1)') "[   1:", INT(2048.0/real(xbin)),",   1:",INT(2048.0/real(ybin)), "]"
    else
       write (detsize,'(a21)') "[   1:2048,   1:2048]"
    endif

    ! Set the principal attributes for the output data set
    if (hasAttribute(inDataSetH, "EXTVER")) then
       int32Value=int32Attribute(inDataSetH, "EXTVER")
       call setAttribute(PPSOSWDataSetH, "EXTVER", int32Value, "Extension Version" )
    endif
    !
    if (hasAttribute(inDataSetH, "TELESCOP")) then
       attrValue=stringAttribute(inDataSetH, "TELESCOP")
       call setAttribute(PPSOSWDataSetH, "TELESCOP", trim(attrValue), "XMM Mission" )
    endif
    !
    if (hasAttribute(inDataSetH, "INSTRUME")) then
       attrValue=stringAttribute(inDataSetH, "INSTRUME")
       call setAttribute(PPSOSWDataSetH, "INSTRUME", trim(attrValue), "Instrument name" )
    endif
    !
    if (hasAttribute(inDataSetH, "CHAIN")) then
       int32Value=int32Attribute(inDataSetH, "CHAIN")
       call setAttribute(PPSOSWDataSetH, "CHAIN", int32Value, "Telemetry chain" )
    endif
    !
    if (hasAttribute(inDataSetH, "DATATYPE")) then
       attrValue=stringAttribute(inDataSetH, "DATATYPE")
       call setAttribute(PPSOSWDataSetH, "DATATYPE", trim(attrValue), "Data type" )
    endif
    !
    if (hasAttribute(inDataSetH, "OBS_ID")) then
       attrValue=stringAttribute(inDataSetH, "OBS_ID")
       call setAttribute(PPSOSWDataSetH, "OBS_ID", trim(attrValue), "Observation Id" )
    endif
    !
    if (hasAttribute(inDataSetH, "EXP_ID")) then
       attrValue=stringAttribute(inDataSetH, "EXP_ID")
       call setAttribute(PPSOSWDataSetH, "EXP_ID", trim(attrValue), "Exposure Id" )
    endif
    !
    if (hasAttribute(inDataSetH, "TMEXP_ID")) then
       int32Value=int32Attribute(inDataSetH, "TMEXP_ID")
       call setAttribute(PPSOSWDataSetH, "TMEXP_ID", int32Value, "Exposure Id from telemetry" )
    endif
    !
    if (hasAttribute(inDataSetH, "SCIWIN")) then
       int32Value=int32Attribute(inDataSetH, "SCIWIN")
       call setAttribute(PPSOSWDataSetH, "SCIWIN", int32Value, "Science window?" )
    endif
    !
    if (hasAttribute(inDataSetH, "PARAM")) then
       int32Value=int32Attribute(inDataSetH, "PARAM")
       call setAttribute(PPSOSWDataSetH, "PARAM", int32Value, "Data Set Parameter" )
    endif
    !
    if (hasAttribute(inDataSetH, "FRMTIME")) then
       int32Value=int32Attribute(inDataSetH, "FRMTIME")
       call setAttribute(PPSOSWDataSetH, "FRMTIME", int32Value, "Frame Time" )
    endif
    !
    if (hasAttribute(inDataSetH, "DATE-OBS")) then
       attrValue=stringAttribute(inDataSetH, "DATE-OBS")
       call setAttribute(PPSOSWDataSetH, "DATE-OBS", trim(attrValue), "Start Time of exposure" )
    endif
    !
    if (hasAttribute(inDataSetH, "DATE-END")) then
       attrValue=stringAttribute(inDataSetH, "DATE-END")
       call setAttribute(PPSOSWDataSetH, "DATE-END", trim(attrValue), "End Time of exposure" )
    endif
    if (hasAttribute(inDataSetH, "BLANK")) then
    !
       int32Value=int32Attribute(inDataSetH, "BLANK")
       call setAttribute(PPSOSWDataSetH, "BLANK", int32Value, "Value indicating missing pixel" )
    endif
    !
    if (hasAttribute(inDataSetH, "BINAX1")) then
       int32Value=int32Attribute(inDataSetH, "BINAX1")
       call setAttribute(PPSOSWDataSetH, "BINAX1", int32Value, "CCD x-axis binning 2**N" )
    endif
    !
    if (hasAttribute(inDataSetH, "BINAX2")) then
       int32Value=int32Attribute(inDataSetH, "BINAX2")
       call setAttribute(PPSOSWDataSetH, "BINAX2", int32Value, "CCD y-axis binning 2**M" )
    endif
    !
    if (hasAttribute(inDataSetH, "EXPOSURE")) then
       int32Value=int32Attribute(inDataSetH, "EXPOSURE")
       call setAttribute(PPSOSWDataSetH, "EXPOSURE", int32Value, "Exposure duration (Secs)")
    endif
    !
    if (hasAttribute(inDataSetH, "BINBPE")) then
       attrValue=stringAttribute(inDataSetH, "BINBPE")
       call setAttribute(PPSOSWDataSetH, "BINBPE", trim(attrValue), "BPE binning enabled" )
    endif
    !
    if (hasAttribute(inDataSetH, "FILTER")) then
       attrValue=stringAttribute(inDataSetH, "FILTER")
       call setAttribute(PPSOSWDataSetH, "FILTER", trim(attrValue), "OM filter" )
    endif
    !
    if (hasAttribute(inDataSetH, "DATAMODE")) then
       attrValue=stringAttribute(inDataSetH, "DATAMODE")
       call setAttribute(PPSOSWDataSetH, "DATAMODE", trim(attrValue), "File content" )
    endif

!----------
    call setAttribute(PPSOSWDataSetH, "DETSIZE", trim(detsize), "Mosaic Keywords")
    call setAttribute(PPSOSWDataSetH, "DETSEC", trim(detSec), "Mosaic Keywords")

    if (hasAttribute(inDataSetH, "FRAMTIME")) then
       real32Value=real32Attribute(inDataSetH, "FRAMTIME")
       call setAttribute(PPSOSWDataSetH, "FRAMTIME", real32Value, "sec", "Frame time")
    endif

    if (hasAttribute(inDataSetH, "DEADFRAC")) then
       real32Value=real32Attribute(inDataSetH, "DEADFRAC")
       call setAttribute(PPSOSWDataSetH, "DEADFRAC", real32Value, "sec", "Frame dead fraction ")
    endif

    outQualityArrayH = addArray(PPSOSWDataSetH, "QUALITY", integer8, dim, "flags", "IMAGE")
    allocate (outQualArray(0 : nxout-1, 0 : nyout-1), stat = allocErr)
    call CheckAllocation(allocErr, "outQualArray")
    outQualArray=0

    blockH=block(inDataSetH, "MODES", read)
    call copyBlock(PPSOSWDataSetH, blockH)

    !********************************************
    ! convert from ra and dec numbers to strings
    !********************************************
 
    call SLA_CR2TF(2, real(ra_point), cs, ihmsf)
    write(rastring, '(i2.2,'':'',i2.2,'':'',i2.2,''.'',i2.2)') &
         ihmsf(1), ihmsf(2), ihmsf(3), ihmsf(4)
 
    call SLA_CR2AF(1, real(dec_point), cs, idmsf)
    write(decstring, '(a1,i2.2,'':'',i2.2,'':'',i2.2,''.'',i1)') &
         cs, idmsf(1), idmsf(2), idmsf(3), idmsf(4)

    !******************************************************************************
    ! now determine refDX, refDY offsets of OM plate centre (PC) from original OSW 
    ! origin (IN BINNED PIXELS)
    !******************************************************************************
    refDX = (xCentre - x0) / dXbinning
    refDY = (yCentre - y0) / dYbinning
!--------Test VNY 21.07.03 (The refx and refy should be unrotated for the undistorted image) 
    refx = trans(1) + refDX
    refy = trans(4) + refDY
    !-VNY 07.11.2003
    if (produceUndistortedImage) then
       call SetFitsArrayAttributes(undistArrayH, xBinning, yBinning, plateScale, refx, refy, &
            ra_scx, dec_scx, ra_point, dec_point)
       undistImageP => real32array2data(undistArrayH)
    else
       allocate (undistImage(0 : nxout-1, 0 : nyout-1), stat = allocErr)
       call CheckAllocation(allocErr, "undistImage")
       undistImage=0
    endif ! produceUndistortedImage

!---
!  : before rotation, the refx and refy should be modified for distortion
    rawx(0) = xCentre       ! position in full OM field.
    rawy(0) = yCentre
    call CAL_omDistortion(rawx, rawy, deltax, deltay)
    refDX=refDX-dble(deltax(0))
    refDY=refDY-dble(deltay(0))
!-------------

    ! Compute these offsets in terms of the rotated frame (remember rotation 
    ! takes place about the OM plate centre so the x,y offsets of the OSW origin 
    ! shift in the rotated frame). 

    rotatedRefDX = refDX * cosRollangle - refDY * sinRollangle
    rotatedRefDY = refDX * sinRollangle + refDY * cosRollangle 

    ! Finally compute offsets (refx, refy) of the OM plate centre from the BB origin, i.e. 
    ! just add the offset of the PC from the OSW origin to the offset of the OSW origin 
    ! from the BB origin

    refx = trans(1) + rotatedRefDX
    refy = trans(4) + rotatedRefDY
    !
    ! And write the WCS attributes to output file. Note that the scale is in binned pixels (ie 
    ! if binning is 2, with raw scale -0.5" per pixels, CDELT1/2 =1
    !
    !****************************************************
    ! Set the fits header attributes for the image array
    !****************************************************
    call SetFitsArrayAttributes(PPSOSWArrayH, xBinning, yBinning, plateScale, refx, refy, &
                                  ra_scx, dec_scx, ra_point, dec_point)
    outImage => real32array2data(PPSOSWArrayH)
    !
    outQualityArrayP => int8Array2Data(outQualityArrayH)
    outQualityArrayP=outQualArray

    !*********************************************************  
    !  Now rotate the image.  
    !*********************************************************
    k = 0
    do i = 0, nxout - 1
        do j = 0, nyout - 1
            call setToNull( PPSOSWArrayH, k)
            k = k + 1
        end do
    end do
    !*********************************************************  
    !  Initialise the undistorted image.  
    !*********************************************************
    if (produceUndistortedImage) then 
       k = 0
       do i = 0, nxdist - 1
          do j = 0, nydist - 1
             call setToNull( undistArrayH, k)
             k = k + 1
          end do
       end do
       call UndistortImage(nx, ny, dXbinning, dYbinning, x0, y0, nxdist, nydist, &
            oswImage, undistImageP)
       call RotateImage(nxdist, nydist, dXbinning, dYbinning, x0, y0, nxout, nyout, &
            undistImageP, outImage, itrans) 
    else
       call UndistortImage(nx, ny, dXbinning, dYbinning, x0, y0, nxdist, nydist, &
            oswImage, undistImage)
       call RotateImage(nxdist, nydist, dXbinning, dYbinning, x0, y0, nxout, nyout, &
            undistImage, outImage, itrans) 
    endif

    if (hasAttribute(inDataSetH,"DETSEC")) then
       window0 = stringAttribute(inDataSetH, "DETSEC")
       read (window0, '(A1, I4, A1, I4, A1, I4, A1, I4, A1)')  &
            ch1,winX0, ch2, winX1, ch3,  winY0, ch4, winY1, ch5

       write (detsec, '(a1,i4,a1,i4,a1,i4,a1,i4,a1)') "[",winX0, ":", winX1, &
            & ",",winY0, ":", winY1, "]" 
       if (xbin .ne. 0. .and. ybin .ne. 0.) then
          write (detsize,'(a6,i4,a6,i4,a1)') "[   1:", INT(2048.0/real(xbin)),",   1:",INT(2048.0/real(ybin)), "]"
       else
          write (detsize,'(a21)') "[   1:2048,   1:2048]"
       endif
       if (produceUndistortedImage)  then
          call setAttribute(undistDataSetH, "DETSIZE", trim(detsize), "Mosaic Keywords")
          call setAttribute(undistDataSetH, "DETSEC", trim(detSec), "Mosaic Keywords")
       endif ! produceUndistortedImage
    else
       write(errstr, '(a)') "DETSEC keyword missing"
       call message(AppMsg, Verbose)
    endif


   ! Release handles

    call release(inArrayH)
    if (hasQUALITY) then
       call release(qualArrayH)
    endif
    call release(PPSOSWArrayH) 
    call release(outQualityArrayH)

    if (produceUndistortedImage) then
       call release(undistArrayH)
       call release(undistDataSetH)
    endif

    call release(PPSOSWDataSetH)
    call release(inDataSetH)

    !**************************************
    ! Release memory allocation
    !**************************************
    if(allocated(qualArray)) then
       deallocate(qualArray, stat=allocErr)
       call CheckDeallocation(allocErr, "qualArray") 
    endif
    if(allocated(undistImage)) then
       deallocate(undistImage, stat=allocErr)
       call CheckDeallocation(allocErr, "undistImage") 
    endif

end subroutine omgprep
!******************************************
! invert_trans
! inverts the transformation matrix, i.e. gets coefficients to swap between
! x'= t(1) + t(2)*x + t(3)*y
! y'= t(4) + t(5)*x + t(6)*y
! and 
! x= it(1) + it(2)*x' + it(3)*y'
! y= it(4) + it(5)*x' + it(6)*y'
!******************************************
subroutine invert_trans(trans, itrans)
    real(kind=double), intent(in) ::  trans(6)
    real(kind=double), intent(out) :: itrans(6)
    real(kind=double) determ

    determ = trans(2) * trans(6) - trans(3) * trans(5)
    itrans(1) = trans(3) * trans(4) - trans(1) * trans(6)
    itrans(2) = trans(6)
    itrans(3) = -trans(3)
    itrans(4) = trans(1) * trans(5) - trans(2) * trans(4)
    itrans(5) = -trans(5)
    itrans(6) = trans(2)
    itrans = itrans / determ

end subroutine invert_trans
!*******************************************
! Function MissionTime - Returns the number
! of seconds from the start of the mission
! to the observation time. Added by CB because
! the call to CAL_GetBoresiteMatrix needs
! the time in seconds from the launch.
!******************************************
double precision function  MissionTime(primaryTable)
    type(DataSetT), intent(in)   :: primaryTable

    character(len = 19)  :: dateObs         
    character(len = 19)  :: dateEnd        
    real(kind = double)  :: JulianDate0, JulianDate1, JulianDate2
    character(len = 19)  :: missionReferenceTime = "1998-01-01T00:00:00"

    ! Compute the Julian date of the mission reference time.
    JulianDate0 = ModifiedJulianDate(missionReferenceTime)

    ! Compute the Julian date of the start of the observations.
    dateObs = stringAttribute(primaryTable,  "DATE-OBS")    
    JulianDate1 = ModifiedJulianDate(dateObs)

    ! Compute the Julian date of the end of the observations.
    dateEnd = stringAttribute(primaryTable,  "DATE-END")   
    JulianDate2 = ModifiedJulianDate(dateEnd)

    ! The required time is the difference between the average of the 
    ! start and end times and the mission reference time, in seconds.
    MissionTime = ((JulianDate1 + JulianDate2) / 2.0d0 - JulianDate0) * 86400.0d0

end function MissionTime
!*******************************************************************************
! subroutine GetImageBoundaries
! The first step is to find the limits of the sky box that will bound the rotated 
! image, after it is linearised (distortion corrections applied). This is done by 
!
! 1) computing the raw coordinates of the image corners and applying the appropriate 
! distortion correction to them. The coordinates are referenced back to the origin 
! of the undistorted image. The corners are then rotated by the roll angle and 
! converted to binned pixels 
! 2) the limits of this rotated, distorted image are obtained and the translational 
! shift betwen the bounding box origin and the OSW origin are passed to the 
! transformation matrix leading coefficients.
! 3) The dimensions of the bounding box are determined
! 4) The offset of the OM plate centre (1024,1024) are determined relative to 
! the OSW original, as measured in the rotated frame, and then this is 
! corrected to be wrt to the bounding box origin
! 5) WCS keywords are set
! 6) pixels in the sky box are now converted, via the inverse transformation matrix, 
! to the OSW frame and an iterative process to correct back from linearised to raw 
! pixels is used to find the original pixel in the raw image. Bi-linear interpolation 
! is used to fill the rotated image  
!*******************************************************************************
subroutine GetImageBoundaries(nx, ny, x0, y0, xBinning, yBinning, &
                          trans, xMin, xMax, yMin, yMax)
    integer, intent(in) :: nx, ny   ! The image dimensions
    real(kind=single),  intent(in) :: x0, y0
    integer, intent(in) :: xBinning, yBinning
    real(kind=double), intent(inout), dimension(1 : 6) :: trans   
    real(kind=double), intent(out) :: xMin, xMax, yMin, yMax
    real(kind=single) :: rawx(0 : 0), rawy(0 : 0)      ! CAL coords
    real(kind=single), pointer     :: deltax(:), deltay(:)       ! distortion offsets
    real(kind=double)  :: xlin(4), ylin(4), xcorn(4), ycorn(4)

! NOTE definition of CAL_omDistortion means we could set up rawx/y arrays 
! first and then call distortion routine once (passing all corners in at once) and then 
! put loop after to compute lin coords (SRR).
     
! bottom left hand corner

    rawx(0) = x0                              !at 2048 pixels resolution 
    rawy(0) = y0
    call CAL_omDistortion(rawx, rawy, deltax, deltay)   
 
    xlin(1) = rawx(0) - deltax(0) - x0            !correct for distortion and reference to x0,y0 origin
    ylin(1) = rawy(0) - deltay(0) - y0
    xcorn(1) = (trans(1) + trans(2) * xlin(1) + trans(3) * ylin(1)) / dble(xBinning)      !rotate + convert to binned pixels
    ycorn(1) = (trans(4) + trans(5) * xlin(1) + trans(6) * ylin(1)) / dble(yBinning)


!bottom rh corner in 2048 pixels
    rawx(0) = x0 + real(nx * xBinning)      
    rawy(0) = y0
    call CAL_omDistortion(rawx, rawy, deltax, deltay)   
    xlin(2) = rawx(0) - deltax(0) - x0
    ylin(2) = rawy(0) - deltay(0) - y0
    xcorn(2) = (trans(1) + trans(2) * xlin(2) + trans(3) * ylin(2)) / dble(xBinning)
    ycorn(2) = (trans(4) + trans(5) * xlin(2) + trans(6) * ylin(2)) / dble(yBinning)

! top rh corner

    rawx(0) = x0 + real(nx * xBinning)
    rawy(0) = y0 + real(ny * yBinning)
    call CAL_omDistortion(rawx, rawy, deltax, deltay)   
    xlin(3) = rawx(0) - deltax(0) - x0               
    ylin(3) = rawy(0) - deltay(0) - y0
    xcorn(3) = (trans(1) + trans(2) * xlin(3) + trans(3) * ylin(3)) / dble(xBinning)
    ycorn(3) = (trans(4) + trans(5) * xlin(3) + trans(6) * ylin(3)) / dble(yBinning)

! top lh corner

    rawx(0) = x0
    rawy(0) = y0 + real(ny * yBinning)
    call CAL_omDistortion(rawx, rawy, deltax, deltay)   
    xlin(4) = rawx(0) - deltax(0) - x0
    ylin(4) = rawy(0) - deltay(0) - y0
    xcorn(4) = (trans(1) + trans(2) * xlin(4) + trans(3) * ylin(4)) / dble(xBinning)
    ycorn(4) = (trans(4) + trans(5) * xlin(4) + trans(6) * ylin(4)) / dble(yBinning)


!  We now have the corners of OSW in linearised coordinates relative to the origin 
!  of the original OSW image, after rotation of the frame 
!  Now need to get limits of bounding box (BB)  from these corners (in rotated frame)
!  Once minimum x and y are found (origin of BB - i.e. the rotated frame), since these 
!  are wrt the original OSW origin, we can determine the offset of the OSW origin from 
!  bounding BB (transfer offsets to trans(1), trans(4))

    xmax = trans(1)
    xmin = trans(1)
    call GetMinMax(xmin, xmax, xcorn)
    trans(1) = trans(1) - xmin                        !get x offset of OSW origin from BB origin

    ymax = trans(4)
    ymin = trans(4)
    call GetMinMax(ymin, ymax, ycorn)
    
    trans(4) = trans(4) - ymin                        !get y offset of OSW origin from BB origin

end subroutine GetImageBoundaries
!******************************
! subrioutine GetMinMax
!******************************
subroutine GetMinMax(minimum, maximum, corner)
    real(kind=double), intent(inout) :: minimum
    real(kind=double), intent(inout) :: maximum
    real(kind=double), intent(in), dimension(1 : 4) :: corner
    integer :: i

    do i = 1, 4
        minimum = dmin1(minimum, corner(i))
        maximum = dmax1(maximum, corner(i))
    end do

end subroutine GetMinMax 
!*****************************************
! subroutine SetFitsArrayAttributes
!*****************************************
subroutine SetFitsArrayAttributes(arrayH, xBinning, yBinning,     &
                                  plateScale, refx, refy, ra_scx, &
                                  dec_scx, ra_point, dec_point)

    type(ArrayT), intent(in) :: arrayH   ! Array handle.
    integer, intent(in) :: xBinning, yBinning 
    real(kind=double), intent(in) :: plateScale
    real(kind=single), intent(in) :: refx, refy
    real(kind=double), intent(in) :: ra_scx, dec_scx
    real(kind=double), intent(in) :: ra_point, dec_point


    call setAttribute(arrayH, "CTYPE1", "RA---TAN", "Coordinate system for X")
    call setAttribute(arrayH, "CRPIX1", refx, "WCS reference pixel") 
    call setAttribute(arrayH, "RA_SCX", ra_scx * radiansToDegrees, &
                              "Coordinate of spacecraft pointing direction for X")
    call setAttribute(arrayH, "CRVAL1", ra_point * radiansToDegrees, &
                              "Coordinate of pointing direction for X")
    call setAttribute(arrayH, "CDELT1", (-1.0)*plateScale*(real(xBinning) * radiansToDegrees), &
                              "WCS pixel size")
    call setAttribute(arrayH, "CTYPE2", "DEC--TAN", "Coordinate system for Y")
    call setAttribute(arrayH, "CRPIX2", refy, "WCS reference pixel")
    call setAttribute(arrayH, "DEC_SCX", dec_scx * radiansToDegrees, &
                              "Coordinate of spacecraft pointing direction for Y")
    call setAttribute(arrayH, "CRVAL2", dec_point * radiansToDegrees, &
                              "Coordinate of pointing direction for Y")
    call setAttribute(arrayH, "CDELT2", plateScale*(real(yBinning)*radiansToDegrees), &
                              "WCS pixel size")
    call setAttribute(arrayH, "RA_PNT", ra_point * radiansToDegrees, &
                              "Coordinate of pointing direction for X")
    call setAttribute(arrayH, "DEC_PNT", dec_point * radiansToDegrees, &
                              "Coordinate of pointing direction for Y")
 
end subroutine SetFitsArrayAttributes


!***************************************************************
! Subroutine RotateImage
! Thius is done by-
! 1) looping over regular grid of output (sky) image points, 
! 2) computing (linear) location in unrotated image (via inverse transform)
! 3) correcting linear position in unrotated frame for distortion (i.e.  
!     converting back to raw pixel position) (2 stage iteration)
! 4) bi-linearly interpolating surrounding points 
!
!****************************************************************
subroutine  RotateImage(nx, ny, dXbinning, dYbinning, x0, y0, & 
                      nxout, nyout, oswImage, rotatedImage, &
                      itrans)  
    integer, intent(in) :: nx, ny
    real(kind=double), intent(in) :: dXbinning, dYbinning
    integer, intent(in)             :: x0, y0             ! offset of image
    integer, intent(in) :: nxout, nyout
    real(kind=single), intent(in), dimension(0 : nx - 1, 0 : ny - 1) :: oswImage
    real(kind=single), intent(out), dimension(0 : nxout - 1, 0 : nyout - 1) :: rotatedImage
    real(kind=double), intent(in)  :: itrans(6)     ! its inverse
  
    !*******************************
    ! Local variables
    !*******************************
    real(kind=double)  :: t, u, rx0, ry0, rx, ry
    real               :: rxInit, ryInit
    real               :: rawx(0 : 0), rawy(0 : 0)  ! CAL coords
    real(kind=double), dimension(:), pointer :: itrans3TimesjPlusPoint5  ! Stores itrans(3) * (j + 0.5)
    real(kind=double), dimension(:), pointer :: itrans6TimesjPlusPoint5  ! Stores itrans(6) * (j + 0.5)
    real, pointer     :: deltax(:), deltay(:)       ! distortion offsets
    real(kind=double) :: dblNx, dblNy
    real(kind=double) :: radius1, radius2, oldDiff, newDiff
    real(kind=double) :: rxOld, ryOld
    integer :: i, j, j0, k0 
    integer :: j1, k1, iter, istat
    integer :: nxMinus2, nyMinus2
    integer :: maxIterations 

!setting a value to all pixels of the rotated image
    rotatedImage=-1.0

    nxMinus2 = nx - 2
    nyMinus2 = ny - 2
    maxIterations = 3

    itrans3TimesjPlusPoint5 => AllocateDouble(0 , nyout - 1, "itrans3TimesjPlusPoint5")
    itrans6TimesjPlusPoint5 => AllocateDouble(0 , nyout - 1, "itrans6TimesjPlusPoint5")
    
    !***************************************************************
    ! The following is to reduce execution time when creating the
    ! rotated image
    !***************************************************************
    forall (j = 0 : nyout - 1) 
        itrans3TimesjPlusPoint5(j) = itrans(3) * (dble(j) + 0.5_double)
        itrans6TimesjPlusPoint5(j) = itrans(6) * (dble(j) + 0.5_double)
    end forall


    dblNx = dble(nx)
    dblNy = dble(ny)

    do i = 0, nxout - 1
       rx0 = itrans(1) + itrans(2) * (dble(i) + 0.5_double)
       ry0 = itrans(4) + itrans(5) * (dble(i) + 0.5_double)
     
       do j = 0, nyout - 1

          !***************************************************************      
          !  use inverse transformation to get coords in unrotated 
          ! (input) frame note part of this is already done within i loop
          !***************************************************************
          rx = rx0 + itrans3TimesjPlusPoint5(j)
          ry = ry0 + itrans6TimesjPlusPoint5(j)


          !***************************************************************************
          !  Calculate position in full frame. Get distortion correction and correct 
          !  positions. THIS is done iteratively (twice) because we want to go from 
          !  linearised to raw position but CAL distortion correction routine gives 
          !  values based on raw input positions. Thus we correct linearised position 
          !  with wrong correction values. To improve the final position we apply 
          !  the wrong distortion correction to get new coordinates and then recompute 
          !  the corrections using this position. This should bring us close to the 
          !  true raw position. 
          !
          ! VNY 10.09.2001 correction for the half of the pixel as the CCD 
          ! grid passes through the pixel center
          !****************************************************************** 
        
          rxInit = rx
          ryInit = ry
          oldDiff = 1e6

          do iter = 1, maxIterations
             rawx(0) = (rx - 0.5) * dXbinning + x0       ! position in full OM field.
             rawy(0) = (ry - 0.5) * dYbinning + y0


             call CAL_omDistortion(rawx, rawy, deltax, deltay)


             rx = rxInit + deltax(0) / dXbinning   ! 25.8.2000 The signs are +ve because we are going from linear 
             ry = ryInit + deltay(0) / dYbinning   ! to raw positions. Convert back to binned pixels.
             ! VNY 19.11.02 The two above lines result in a shift in x-direction
             ! somwhere in the middle of the image (probably distortion is wrong)

!&&&&&&&&&&&&&&&&&&&&&&&&&&&&&& VNY 19.11.2002
             rx=rxInit
             ry=ryInit
!&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&

             radius1 = sqrt(rx * rx + ry * ry)
             radius2 = sqrt(rxInit * rxInit + ryInit * ryInit)
             newDiff = abs(radius1 - radius2)


             if (newDiff .gt. oldDiff) then
                 rx = rxOld
                 ry = ryOld

                 exit
             else  
                 if (newDiff  .lt. 0.05) then
                 
                     exit
                 
                 endif 
             end if
             oldDiff = newDiff
             rxOld = rx
             ryOld = ry
          end do

          !*********************************************
          ! use bi-linear interpolation to fill in grid
          !*********************************************
          if (rx .ge. 0.0_double .and. rx .le. dblNx .and. & 
              ry .ge. 0.0_double .and. ry .le. dblNy) then

              !********************************************************************
              !   VNY 10.09.2001 correction for the half of the pixel because
              !   the origin of the CCD coordinate frame is in the pixel centre 
              !   
              !********************************************************************
              rx = rx - 0.5_double
              ry = ry - 0.5_double

              j0 = max(0, int(rx))
              j0 = min(nxMinus2, j0)
              k0 = min(nyMinus2, int(ry))
              k0 = max(0, k0)
              j1 = j0 + 1
              k1 = k0 + 1

              t = (rx - dble(j0))  
              u = (ry - dble(k0))

              rotatedImage(i, j ) = (1.0_double - t) * (1.0_double - u) * oswImage(j0, k0) + &
                                    t                * (1.0_double - u) * oswImage(j1, k0) + &
                                    t                * u                * oswImage(j1, k1) + &
                                   (1.0_double - t)  * u                * oswImage(j0, k1)

          end if
       end do
    end do

    deallocate(itrans3TimesjPlusPoint5, stat = istat)
    call  CheckDeAllocation(istat, "itrans3TimesjPlusPoint5")
 
    deallocate(itrans6TimesjPlusPoint5, stat = istat)
    call  CheckDEAllocation(istat, "itrans6TimesjPlusPoint5")
 

end subroutine RotateImage

!***************************************************************
! Subroutine UndistortImage
! Thius is done by-
! 1) looping over regular grid of output (sky) image points, 
! 2) computing (linear) location in unrotated image (via inverse transform)
! 3) correcting linear position in unrotated frame for distortion (i.e.  
!     converting back to raw pixel position) (2 stage iteration)
! 4) bi-linearly interpolating surrounding points 
!
!****************************************************************
subroutine  UndistortImage(nx, ny, dXbinning, dYbinning, x0, y0, & 
                      nxout, nyout, oswImage, undistortedImage)
    integer, intent(in) :: nx, ny
    real(kind=double), intent(in) :: dXbinning, dYbinning
    integer, intent(in)             :: x0, y0             ! offset of image
    integer, intent(in) :: nxout, nyout
    real(kind=single), intent(in), dimension(0 : nx - 1, 0 : ny - 1) :: oswImage
    real(kind=single), intent(out), dimension(0 : nxout - 1, 0 : nyout - 1) :: undistortedImage
  
    !*******************************
    ! Local variables
    !*******************************
    real(kind=double)  :: rx, ry!, rx0, ry0, t, u 
    real               :: rxNew, ryNew, di, dj
    real               :: rawx(0 : 0), rawy(0 : 0)  ! CAL coords
    real, pointer     :: deltax(:), deltay(:)       ! distortion offsets
    real(kind=double) :: dblNx, dblNy
    real(kind=double) :: dblNxOut, dblNyOut
!    real(kind=double) :: radius2, oldDiff, newDiff, radius1
!    real(kind=double) :: rxOld, ryOld
    integer :: i, j!, k0, j0
    integer :: iout, jout
    integer :: i1, j1!, istat, iter, k1
    integer :: nxMinus2, nyMinus2
    integer :: maxIterations 
    real(kind=double) :: totImage, leftImage, rightImage, leftUpImage, leftDownImage, rightUpImage, rightDownImage
!---VNY Test 19.11.2002
!setting a value to all pixels of the rotated image
    undistortedImage=0.0

    nxMinus2 = nx - 2
    nyMinus2 = ny - 2
    maxIterations = 3

    
    dblNx = dble(nx)
    dblNy = dble(ny)
! VNY 22.07.03
    dblNxOut = dble(nxout)
    dblNyOut = dble(nyout)
    do i = 0, nx - 1
       rx = dble(i) !+ 0.5_double
       do j = 0, ny - 1
          ry = dble(j) !+ 0.5_double 
          rawx(0) = (rx + x0)*dXbinning       ! position in full OM field.
          rawy(0) = (ry + y0)*dYbinning
          call CAL_omDistortion(rawx, rawy, deltax, deltay)

!          rxInit = rx - deltax(0) / dXbinning   ! 25.8.2000 The signs are +ve because we are going from linear 
!          ryInit = ry - deltay(0) / dYbinning   ! to raw positions. Convert back to binned pixels.
          rxNew = rx - deltax(0) /dXbinning     ! 25.8.2000 The signs are +ve because we are going from linear 
          ryNew = ry - deltay(0) /dYbinning   ! to raw positions. Convert back to binned pixels.
!---
!          rxNew = rx - deltax(0) /dXbinning     ! 25.8.2000 The signs are +ve because we are going from linear 
!          ryNew = ry - deltay(0) /dYbinning   ! to raw positions. Convert back to binned pixels.
!---
          iout=int(rxNew)
          jout=int(ryNew)
          di=rxNew-real(iout)
          dj=ryNew-real(jout)
          i1=1
          if (di .lt. 0) then
             i1=-1
             di = -di
          endif
          j1=1
          if (dj .lt. 0) then
             j1=-1
             dj=-dj
          endif


          ! Subdivision of the image in four parts
          totImage=oswImage(i,j)
          leftImage=totImage*di+0.5
          rightImage=totImage-leftImage
          leftDownImage=leftImage*dj+0.5
          leftUpImage=leftImage-leftDownImage
          rightDownImage=rightImage*dj+0.5
          rightUpImage=rightImage-rightDownImage

!---
          if (iout .lt. nxout .and. iout .ge. 0 .and. jout .lt. nyout .and. jout .ge. 0)  then
             undistortedImage(iout,jout)= undistortedImage(iout,jout)+rightUpImage
          endif
          if (iout+i1 .lt. nxout .and. iout+i1 .ge. 0 .and. jout .lt. nyout .and. jout .ge. 0)  then
             undistortedImage(iout+i1,jout)= undistortedImage(iout+i1,jout)+leftUpImage
          endif
          if (iout .lt. nxout .and. iout .ge. 0 .and. jout+j1 .lt. nyout .and. jout+j1 .ge. 0)  then
             undistortedImage(iout,jout+j1)= undistortedImage(iout,jout+j1)+rightDownImage
          endif
          if (iout+i1 .lt. nxout .and. iout+i1 .ge. 0 .and. jout+j1 .lt. nyout .and. jout+j1 .ge. 0)  then
             undistortedImage(iout+i1,jout+j1)= undistortedImage(iout+i1,jout+j1)+leftDownImage
          endif
       end do
    end do
end subroutine UndistortImage

!********************************************
! Subroutine ModifySourcePositionAngles
!*******************************************
subroutine ModifySourcePositionAngles(tableSetH, angle, nullValue)
    type(TableT), intent(in) :: tableSetH
    real(kind=double), intent(in) :: angle
    real(kind=single), intent(in) :: nullValue
    
    real(kind=single), pointer :: positionAngle(:)
    integer(kind=int32) :: noRows, position
  
    noRows = numberOfRows(tableSetH)

    position = columnNumber(tableSetH, "PA")

    positionAngle => AllocateSingle(1, noRows, "positionAngle")
    call ReadReal32Column(tableSetH, "PA", noRows, 0, positionAngle)
 
    call deleteColumn(tableSetH, "PA")

    positionAngle = positionAngle + angle
    call AddReal32Column(tableSetH, positionAngle, noRows, "PA", "arcsecs", &
             "PA of major-axis Referred to equatorial system", nullvalue, position)
 

    deallocate(positionAngle)
end subroutine ModifySourcePositionAngles


!********************************************************
! subroutine smoothIMage
! Routine to smooth image 
!
!********************************************************
!subroutine SmoothImage1(image, nx, ny, XsmoothSize, YsmoothSize)
subroutine SmoothImage(image, nx, ny, XsmoothSize, YsmoothSize)
 
    integer (int32), intent(in)    :: nx, ny
    real (single), pointer      :: image(:, :)
    

    real (single), pointer   :: tempImage(:, :)
    integer (int32), intent(in)    :: XSmoothSize, ySmoothSize

    ! internal variables
    integer (int32)                :: height
    integer (int32)                :: i

    ! For top hat
    real (single), pointer         :: scan(:)
    integer (int32), pointer       :: nscan(:)
    integer (int32)                :: allocErr

    tempImage => Allocate2DSingle(0, nx - 1, 0 , ny - 1, "tempImage")
     

    !*************************
    ! for gaussian smooth 
    !************************
    height = YSmoothSize
    
    scan => AllocateSingle(-height, height, "scan (smoothImage)")
    nscan => AllocateInt32(-height, height, "nscan (smoothImage)")
   

    !**************************************************
    ! smoothTopHat deals with one X position at a time
    !**************************************************
    do i = 0, nx - 1
        call MovingBox(image, tempImage, nx, ny, XSmoothSize, YSmoothSize, i, scan, nscan, 1)
    end do
  
    deallocate(scan, stat = allocErr)
    call CheckDeallocation(allocErr, "scan")

    deallocate(nscan, stat = allocErr)
    call CheckDeallocation(allocErr, "nscan")
    image = tempImage
    deallocate(tempImage)

!end subroutine SmoothImage1
end subroutine SmoothImage

!******************************************************************************
! subroutine MovingBox
! Top hat smoothing routine - does a strip in Y direction (best array access)
! 12.12.2001 - Modified to use the level image (ie to ignore non-background
! pixels)
!******************************************************************************
subroutine MovingBox(image, backgroundTot, nx, ny, width, height, startX, scan, &
                     nscan, type)

    integer (int32), intent(in)  :: nx, ny
    real (single), pointer    :: image(:, :)

   
    real (single), pointer :: backgroundTot(:, :)


    integer (int32), intent(in) :: width, height     ! Size of smoothing area
    integer (int32), intent(in) :: startX           ! starting point
    real (single), pointer :: scan(:)
    integer (int32), pointer :: nscan(:)

    integer (int32), intent(in) :: type
    !***********************
    ! internal variables
    ! for top hat smooth
    !***********************
    integer (int32)            :: i, j ,k
    real(single)               :: running_total
    integer (int32)            :: nrunning_total
    integer (int32)            :: minXpos, maxXpos
    integer (int32)            :: minYpos, maxYpos
    integer (int32)            :: clearStart, clearEnd
    integer (int32)            :: clearEndPlus1

    real (single)              :: scanval

    !************************************
    ! initialise arrays for this strip
    !************************************
    scan = 0
    nscan = 0

    ! X min and max's are fixed
     
    minXpos = MAX(startX - width, 0)
    maxXpos = MIN(startX + width, nx - 1)
 
    !**************************************************************
    ! Get start and end of clear (ie. don't need MIN/MAX checks )
    !**************************************************************
    clearStart = MIN(height + 1, ny)

    if(height .gt. ny) then
        clearEnd = -1
    else
        clearEnd = ny - height - 1
    end if

    !******************************************
    ! At start need to build up running total  
    ! First one
    !******************************************
    minYpos = 0
    maxYpos = MIN(height, ny - 1)


    running_total= 0.0
    nrunning_total= 0

    do j = minYpos, maxYpos     
        scanval = 0.0
        do k = minXpos, maxXpos 
            scanval = scanval + image(k, j)
        end do
        scan(j) = scanval
        nscan(j) = maxXpos - minXpos + 1
        running_total = running_total + scanval
        nrunning_total = nrunning_total + nscan(j)

    end do
  
    if(type .eq. 1) then
        backgroundTot(startX, 0) = running_total / nrunning_total
    else
        backgroundTot(startX, 0) = running_total
    end if

  
    do i = 1, clearStart - 1
     
        ! Shift down arrays
     
        do j = minYpos,  maxYpos
            scan(j - 1) = scan(j)
            nscan(j - 1) = nscan(j)
        end do
     
        ! reset min/max values
     
        minYpos = MAX(i - height, 0)
        maxYpos = MIN(i + height, ny - 1)

        !***********************
        ! Add in new entry
        !***********************     
        scanval = 0.0
        do k = minXpos, maxXpos 
            scanval = scanval + image(k, maxYpos)
        end do
        minYpos = minYpos - i
        maxYpos = maxYpos - i
        scan(maxYpos) = scanval
        nscan(maxYpos) = maxXpos - minXpos + 1
        running_total = running_total + scanval
        nrunning_total = nrunning_total + nscan(maxYpos)
 
        if(type .eq. 1) then
            backgroundTot(startX, i) = running_total / nrunning_total
        else
            backgroundTot(startX, i) = running_total
        end if
    
    end do
 

    !********************************
    ! Now do the rest (not MIN/MAX)
    !********************************
    if(clearEnd .eq. -1) return
  

    ! Make sure we can hold whole smoothing area within data

    if(clearStart .lt. clearEnd) then
        do i = clearStart, clearEnd
     
            ! Shift down arrays and remove bottom line

            running_total = running_total - scan(-height)
            nrunning_total = nrunning_total - nscan(-height)
            do j = -height + 1, height
                scan(j - 1) = scan(j)
                nscan(j - 1) = nscan(j)
            end do
            !*******************
            ! Add in new entry
            !******************
        
            maxYpos = i + height
            scanval = 0.0
            do k = minXpos, maxXpos 
                scanval = scanval + image(k, maxYpos)
            end do

            scan(height) = scanval
            nscan(height) = maxXpos - minXpos + 1
            running_total = running_total + scanval
            nrunning_total = nrunning_total + nscan(height)
       
        
            if(type .eq. 1) then
                backgroundTot(startX, i) = running_total / nrunning_total
            else
                backgroundTot(startX, i) = running_total
            end if

        end do
    end if

    !**********************
    ! And now do end part
    !**********************
    clearEndPlus1 = clearEnd + 1  

    do i = clearEndPlus1, ny - 1
        running_total = running_total - scan(height - i + clearEndPlus1)
        nrunning_total = nrunning_total - nscan(height - i + clearEndPlus1)

        if(type .eq. 1) then
            backgroundTot(startX, i) = running_total / nrunning_total
        else
            backgroundTot(startX, i) = running_total
        end if
    end do

end subroutine MovingBox


end module omgprep_mod





























































































































































































































































